@@ -60,13 +60,12 @@ else:
                 gdkwin.set_cursor(cursor)
 
 import sys
-import uuid
 import os
 import time
 import ctypes
 try:
     from queue import Queue     #@UnresolvedImport @UnusedImport (python3)
-except:
+except ImportError:
     from Queue import Queue     #@Reimport
 
 
@@ -81,19 +80,23 @@ from xpra.deque import maxdeque
 from xpra.client_base import XpraClientBase, EXIT_TIMEOUT
 from xpra.keys import DEFAULT_MODIFIER_MEANINGS, DEFAULT_MODIFIER_NUISANCE, DEFAULT_MODIFIER_IGNORE_KEYNAMES
 from xpra.platform.gui import ClientExtras
-from xpra.scripts.main import ENCODINGS, get_codecs
+from xpra.scripts.config import HAS_SOUND, ENCODINGS, get_codecs
 from xpra.version_util import add_gtk_version_info
 from xpra.stats.base import std_unit
 from xpra.protocol import Compressed
 from xpra.daemon_thread import make_daemon_thread
-
 from xpra.client_window import ClientWindow, DRAW_DEBUG
+from xpra.gtk_util import set_application_name
+
 
 def nn(x):
     if x is None:
         return  ""
     return x
 
+FAKE_BROKEN_CONNECTION = os.environ.get("XPRA_FAKE_BROKEN_CONNECTION", "0")=="1"
+PING_TIMEOUT = int(os.environ.get("XPRA_PING_TIMEOUT", "60"))
+
 
 class XpraClient(XpraClientBase, gobject.GObject):
     __gsignals__ = {
@@ -133,10 +136,12 @@ class XpraClient(XpraClientBase, gobject.GObject):
         self.server_max_desktop_size = None
         self.server_display = None
         self.server_randr = False
+        self.server_auto_refresh_delay = 0
         self.pixel_counter = maxdeque(maxlen=100)
         self.server_ping_latency = maxdeque(maxlen=100)
         self.server_load = None
         self.client_ping_latency = maxdeque(maxlen=100)
+        self._server_ok = True
         self.last_ping_echoed_time = 0
         self.server_info_request = False
         self.server_last_info = None
@@ -172,7 +177,7 @@ class XpraClient(XpraClientBase, gobject.GObject):
         self.mmap_size = 0
 
         #features:
-        self.init_opengl()
+        self.init_opengl(opts.opengl)
         self.toggle_cursors_bell_notify = False
         self.toggle_keyboard_sync = False
         self.window_configure = False
@@ -180,7 +185,7 @@ class XpraClient(XpraClientBase, gobject.GObject):
         self.change_min_quality = False
         self.change_speed = False
         self.readonly = opts.readonly
-        self.windows_enabled = opts.windows_enabled
+        self.windows_enabled = opts.windows
         self._client_extras = ClientExtras(self, opts, conn)
         self.client_supports_notifications = opts.notifications and self._client_extras.can_notify()
         self.client_supports_system_tray = opts.system_tray and self._client_extras.supports_system_tray()
@@ -211,7 +216,7 @@ class XpraClient(XpraClientBase, gobject.GObject):
         except:
             self._keymap = None
         self._do_keys_changed()
-        self.key_shortcuts = self.parse_shortcuts(opts.key_shortcuts)
+        self.key_shortcuts = self.parse_shortcuts(opts.key_shortcut)
         self.send_hello()
 
         if self._keymap:
@@ -228,20 +233,21 @@ class XpraClient(XpraClientBase, gobject.GObject):
                 q -= 10
             elif bw < self.max_bandwidth:
                 q += 5
-            q = max(10, min(95 ,q))
-            self.send_min_quality(q)
+            self.min_quality = max(10, min(95 ,q))
+            self.send_min_quality()
             return True
         if (self.max_bandwidth):
             self.last_input_bytecount = 0
             gobject.timeout_add(2000, compute_receive_bandwidth, 2000)
-        if opts.send_pings:
+        if opts.pings:
             gobject.timeout_add(1000, self.send_ping)
         else:
-            gobject.timeout_add(20*1000, self.send_ping)
+            gobject.timeout_add(10*1000, self.send_ping)
 
     def init_mmap(self, mmap_group, socket_filename):
         log("init_mmap(%s, %s)", mmap_group, socket_filename)
         try:
+            from xpra.platform.uuid_wrapper import get_int_uuid
             import mmap
             import tempfile
             from stat import S_IRUSR,S_IWUSR,S_IRGRP,S_IWGRP
@@ -268,7 +274,7 @@ class XpraClient(XpraClientBase, gobject.GObject):
             os.lseek(fd, 0, SEEK_SET)
             self.mmap = mmap.mmap(fd, length=self.mmap_size)
             #write the 16 byte token one byte at a time - no endianness
-            self.mmap_token = uuid.uuid4().int
+            self.mmap_token = get_int_uuid()
             log.debug("mmap_token=%s", self.mmap_token)
             v = self.mmap_token
             for i in range(0,16):
@@ -284,27 +290,44 @@ class XpraClient(XpraClientBase, gobject.GObject):
             self.mmap_file = None
             self.mmap_size = 0
 
-    def init_opengl(self):
-        USE_OPENGL = os.environ.get("XPRA_OPENGL", "0")=="1"
+    def init_opengl(self, enable_opengl):
+        #enable_opengl can be True, False or None (auto-detect)
         self.opengl_enabled = False
         self.GLClientWindowClass = None
         self.opengl_props = {}
-        #the GL backend only works with gtk2:
-        if USE_OPENGL and not is_gtk3():
+        from xpra.scripts.config import OpenGL_safety_check
+        check = OpenGL_safety_check()
+        if check:
+            if enable_opengl is True:
+                log.warn("OpenGL enabled despite: %s", check)
+            else:
+                self.opengl_props["info"] = "disabled: %s" % check
+                log.warn("OpenGL disabled: %s", check)
+                return
+        if enable_opengl is False:
+            self.opengl_props["info"] = "disabled by configuration"
+            return
+        if is_gtk3():
+            self.opengl_props["info"] = "GTK3 does not support OpenGL"
+            return
+        self.opengl_props["info"] = ""
+        try:
             try:
                 from xpra import gl     #@UnusedImport
-                try:
-                    from xpra.gl.gl_check import check_support
-                    self.opengl_props = check_support()
+                from xpra.gl.gl_check import check_support
+                w, h = get_root_size()
+                min_texture_size = max(w, h)
+                self.opengl_props = check_support(min_texture_size, force_enable=(enable_opengl is True))
 
-                    from xpra.gl.gl_client_window import GLClientWindow
-                    self.GLClientWindowClass = GLClientWindow
-                    self.opengl_enabled = True
-                except Exception, e:
-                    log.error("Error loading OpenGL support: %s", e, exc_info=True)
+                from xpra.gl.gl_client_window import GLClientWindow
+                self.GLClientWindowClass = GLClientWindow
+                self.opengl_enabled = True
             except ImportError, e:
-                log("OpenGL support not installed: %s", e)
-        return self.GLClientWindowClass is not None
+                log.info("OpenGL support not enabled: %s", e)
+                self.opengl_props["info"] = str(e)
+        except Exception, e:
+            log.error("Error loading OpenGL support: %s", e, exc_info=True)
+            self.opengl_props["info"] = str(e)
 
     def init_packet_handlers(self):
         XpraClientBase.init_packet_handlers(self)
@@ -339,16 +362,25 @@ class XpraClient(XpraClientBase, gobject.GObject):
     def run(self):
         gtk_main_quit_on_fatal_exceptions_enable()
         gtk.main()
+        log("XpraClient.run() main loop ended, returning exit_code=%s", self.exit_code)
         return  self.exit_code
 
     def quit(self, exit_code=0):
+        log("XpraClient.quit(%s) current exit_code=%s", exit_code, self.exit_code)
         if self.exit_code is None:
             self.exit_code = exit_code
+        if gtk.main_level()>0:
+            #if for some reason cleanup() hangs, maybe this will fire...
+            gobject.timeout_add(4*1000, gtk_main_quit_really)
+            #try harder!:
+            gobject.timeout_add(5*1000, os._exit, 1)
         self.cleanup()
-        gobject.timeout_add(50, gtk_main_quit_really)
+        if gtk.main_level()>0:
+            log("XpraClient.quit(%s) main loop at level %s, calling gtk quit via timeout", exit_code, gtk.main_level())
+            gobject.timeout_add(500, gtk_main_quit_really)
 
     def cleanup(self):
-        log("cleanup() client_extras=%s", self._client_extras)
+        log("XpraClient.cleanup() client_extras=%s", self._client_extras)
         if self._client_extras:
             self._client_extras.cleanup()
         if self.sound_sink:
@@ -357,8 +389,19 @@ class XpraClient(XpraClientBase, gobject.GObject):
             self.stop_sending_sound()
         XpraClientBase.cleanup(self)
         self.clean_mmap()
+        #the protocol has been closed, it is now safe to close all the windows:
+        #(cleaner and needed when we run embedded in the client launcher)
+        for w in self._id_to_window.values():
+            try:
+                w.destroy()
+            except:
+                pass
+        self._id_to_window = {}
+        self._window_to_id = {}
+        log("XpraClient.cleanup() done")
 
     def clean_mmap(self):
+        log("XpraClient.clean_mmap() mmap_file=%s", self.mmap_file)
         if self.mmap_file and os.path.exists(self.mmap_file):
             os.unlink(self.mmap_file)
             self.mmap_file = None
@@ -440,6 +483,8 @@ class XpraClient(XpraClientBase, gobject.GObject):
             return  True
         try:
             method()
+        except KeyboardInterrupt:
+            raise
         except Exception, e:
             log.error("key_handled_as_shortcut(%s,%s,%s,%s) failed to execute shortcut=%s: %s", window, key_name, modifiers, depressed, shortcut, e)
         return  True
@@ -626,6 +671,11 @@ class XpraClient(XpraClientBase, gobject.GObject):
         capabilities["compressible_cursors"] = True
         capabilities["dpi"] = self.dpi
         capabilities["clipboard"] = self.client_supports_clipboard
+        capabilities["clipboard.notifications"] = self.client_supports_clipboard
+        #buggy osx clipboards:
+        capabilities["clipboard.want_targets"] = sys.platform.startswith("darwin")
+        #buggy osx and win32 clipboards:
+        capabilities["clipboard.greedy"] = sys.platform.startswith("win") or sys.platform.startswith("darwin")
         capabilities["notifications"] = self.client_supports_notifications
         capabilities["cursors"] = self.client_supports_cursors
         capabilities["bell"] = self.client_supports_bell
@@ -667,7 +717,6 @@ class XpraClient(XpraClientBase, gobject.GObject):
             capabilities["encoding.supports_delta"] = []    #need implementing in window_backing
         else:
             capabilities["encoding.supports_delta"] = [x for x in ("png", "rgb24") if x in ENCODINGS]
-        from xpra.scripts.main import HAS_SOUND
         if HAS_SOUND:
             try:
                 from xpra.sound.pulseaudio_util import add_pulseaudio_capabilities
@@ -689,19 +738,53 @@ class XpraClient(XpraClientBase, gobject.GObject):
         log("batch props=%s", [("%s=%s" % (k,v)) for k,v in capabilities.items() if k.startswith("batch.")])
         return capabilities
 
+    def server_ok(self):
+        return self._server_ok
+
+    def check_server_echo(self, ping_sent_time):
+        last = self._server_ok
+        self._server_ok = not FAKE_BROKEN_CONNECTION and self.last_ping_echoed_time>=ping_sent_time
+        if last!=self._server_ok and not self._server_ok:
+            log.info("check_server_echo: server is not responding, redrawing spinners over the windows")
+            def timer_redraw():
+                self.redraw_spinners()
+                if self.server_ok():
+                    log.info("check_server_echo: server is OK again, stopping redraw")
+                    return False
+                return True
+            self.redraw_spinners()
+            gobject.timeout_add(100, timer_redraw)
+        return False
+
+    def redraw_spinners(self):
+        #draws spinner on top of the window, or not (plain repaint)
+        #depending on whether the server is ok or not
+        for w in self._id_to_window.values():
+            if not w.is_tray():
+                w.spinner(self.server_ok())
+
+    def check_echo_timeout(self, ping_time):
+        log("check_echo_timeout(%s) last_ping_echoed_time=%s", ping_time, self.last_ping_echoed_time)
+        if self.last_ping_echoed_time<ping_time:
+            self.warn_and_quit(EXIT_TIMEOUT, "server ping timeout - waited %s seconds without a response" % PING_TIMEOUT)
+
     def send_ping(self):
-        now_ms = int(1000*time.time())
+        now_ms = int(1000.0*time.time())
         self.send("ping", now_ms)
-        wait = 60
-        def check_echo_received(*args):
-            if self.last_ping_echoed_time<now_ms:
-                self.warn_and_quit(EXIT_TIMEOUT, "server ping timeout - waited %s seconds without a response" % wait)
-        gobject.timeout_add(wait*1000, check_echo_received)
+        gobject.timeout_add(PING_TIMEOUT*1000, self.check_echo_timeout, now_ms)
+        wait = 2.0
+        if len(self.server_ping_latency)>0:
+            l = [x for _,x in list(self.server_ping_latency)]
+            avg = sum(l) / len(l)
+            wait = 1.0+avg*2.0
+            log("average server latency=%.1f, using max wait %.2fs", 1000.0*avg, wait)
+        gobject.timeout_add(int(1000.0*wait), self.check_server_echo, now_ms)
         return True
 
     def _process_ping_echo(self, packet):
         echoedtime, l1, l2, l3, cl = packet[1:6]
         self.last_ping_echoed_time = echoedtime
+        self.check_server_echo(0)
         server_ping_latency = time.time()-echoedtime/1000.0
         self.server_ping_latency.append((time.time(), server_ping_latency))
         self.server_load = l1, l2, l3
@@ -711,11 +794,13 @@ class XpraClient(XpraClientBase, gobject.GObject):
 
     def _process_ping(self, packet):
         echotime = packet[1]
-        try:
-            (fl1, fl2, fl3) = os.getloadavg()
-            l1,l2,l3 = int(fl1*1000), int(fl2*1000), int(fl3*1000)
-        except:
-            l1,l2,l3 = 0,0,0
+        l1,l2,l3 = 0,0,0
+        if os.name=="posix":
+            try:
+                (fl1, fl2, fl3) = os.getloadavg()
+                l1,l2,l3 = int(fl1*1000), int(fl2*1000), int(fl3*1000)
+            except (OSError, AttributeError):
+                pass
         sl = -1
         if len(self.server_ping_latency)>0:
             _, sl = self.server_ping_latency[-1]
@@ -732,24 +817,30 @@ class XpraClient(XpraClientBase, gobject.GObject):
             self.info_request_pending = True
             self.send("info-request", [self.uuid], self._id_to_window.keys())
 
-    def send_min_quality(self, q):
+    def send_min_quality(self):
+        q = self.min_quality
         assert q==-1 or (q>=0 and q<=100), "invalid quality: %s" % q
-        self.min_quality = q
         if self.change_min_quality:
             #v0.8 onwards: set min
-            self.send("min-quality", self.min_quality)
+            self.send("min-quality", q)
         elif self.change_quality:
             #v0.7 and earlier, can only set fixed quality..
-            self.send("quality", self.min_quality)
+            self.send("quality", q)
         else:
             #this is really old..
-            self.send("jpeg-quality", self.min_quality)
+            self.send("jpeg-quality", q)
 
-    def send_speed(self, s):
+    def send_speed(self):
         assert self.change_speed
+        s = self.speed
         assert s==-1 or (s>=0 and s<=100), "invalid speed: %s" % s
-        self.speed = s
-        self.send("speed", self.speed)
+        self.send("speed", s)
+
+    def send_min_speed(self):
+        assert self.change_speed
+        s = self.min_speed
+        assert s==-1 or (s>=0 and s<=100), "invalid speed: %s" % s
+        self.send("min-speed", s)
 
     def send_refresh(self, wid):
         self.send("buffer-refresh", wid, True, 95)
@@ -763,6 +854,7 @@ class XpraClient(XpraClientBase, gobject.GObject):
             return
         if not self.session_name:
             self.session_name = capabilities.get("session_name", "Xpra")
+        set_application_name(self.session_name)
         self.window_configure = capabilities.get("window_configure", False)
         self.server_supports_notifications = capabilities.get("notifications", False)
         self.notifications_enabled = self.server_supports_notifications and self.client_supports_notifications
@@ -830,15 +922,6 @@ class XpraClient(XpraClientBase, gobject.GObject):
         if modifier_keycodes:
             self._client_extras.set_modifier_mappings(modifier_keycodes)
 
-        if sys.version_info[:2]<(2,5):
-            log.warn("Python %s is too old!", sys.version_info)
-        else:
-            try:
-                import glib
-                glib.set_application_name(self.session_name)
-            except ImportError, e:
-                log.warn("glib is missing, cannot set the application name, please install glib's python bindings: %s", e)
-
         #sound:
         self.server_pulseaudio_id = capabilities.get("sound.pulseaudio.id")
         self.server_pulseaudio_server = capabilities.get("sound.pulseaudio.server")
@@ -852,10 +935,10 @@ class XpraClient(XpraClientBase, gobject.GObject):
         #if self.server_sound_receive and self.microphone_allowed:
         #    self.start_sending_sound()
 
-        #ui may want to know this is now set:
-        self.emit("clipboard-toggled")
         self.key_repeat_delay, self.key_repeat_interval = capabilities.get("key_repeat", (-1,-1))
         self.emit("handshake-complete")
+        #ui may want to know this is now set:
+        self.emit("clipboard-toggled")
         if self.server_supports_clipboard:
             #from now on, we will send a message to the server whenever the clipboard flag changes:
             self.connect("clipboard-toggled", self.send_clipboard_enabled_status)
@@ -885,7 +968,7 @@ class XpraClient(XpraClientBase, gobject.GObject):
             self.emit("microphone-changed")
         try:
             from xpra.sound.gstreamer_util import start_sending_sound
-            self.sound_source = start_sending_sound(self.server_sound_decoders, self.microphone_codecs, self.server_pulsesound_server, self.server_pulseaudio_id)
+            self.sound_source = start_sending_sound(self.server_sound_decoders, self.microphone_codecs, self.server_pulseaudio_server, self.server_pulseaudio_id)
             if not self.sound_source:
                 return False
             self.sound_source.connect("new-buffer", self.new_sound_buffer)
@@ -899,12 +982,15 @@ class XpraClient(XpraClientBase, gobject.GObject):
 
     def stop_sending_sound(self):
         """ stop the sound source and emit client signal """
+        log("XpraClient.stop_sending_sound()")
         if self.sound_source is None:
             log.warn("stop_sending_sound: sound not started!")
             return
         self.microphone_enabled = False
-        self.sound_source.stop()
+        self.sound_source.cleanup()
+        self.sound_source = None
         self.emit("microphone-changed")
+        log("XpraClient.stop_sending_sound() done")
 
     def start_receiving_sound(self):
         """ ask the server to start sending sound and emit the client signal """
@@ -919,13 +1005,16 @@ class XpraClient(XpraClientBase, gobject.GObject):
 
     def stop_receiving_sound(self):
         """ ask the server to stop sending sound, toggle flag so we ignore further packets and emit client signal """
+        log("XpraClient.stop_receiving_sound()")
         self.send("sound-control", "stop")
         if self.sound_sink is None:
             log("stop_receiving_sound: sound not started!")
             return
         self.speaker_enabled = False
-        self.sound_sink.stop()
+        self.sound_sink.cleanup()
+        self.sound_sink = None
         self.emit("speaker-changed")
+        log("XpraClient.stop_receiving_sound() done")
 
 
     def start_sound_sink(self, codec):
@@ -934,12 +1023,17 @@ class XpraClient(XpraClientBase, gobject.GObject):
             self.emit("speaker-changed")
         def sound_sink_bitrate_changed(*args):
             self.emit("speaker-changed")
+        def sound_sink_error(*args):
+            log.warn("stopping sound because of error")
+            self.stop_receiving_sound()
+            self.emit("speaker-changed")
         try:
             log("starting %s sound sink", codec)
             from xpra.sound.sink import SoundSink
             self.sound_sink = SoundSink(codec=codec)
             self.sound_sink.connect("state-changed", sound_sink_state_changed)
             self.sound_sink.connect("bitrate-changed", sound_sink_bitrate_changed)
+            self.sound_sink.connect("error", sound_sink_error)
             self.sound_sink.start()
             log("%s sound sink started", codec)
             return True
@@ -947,9 +1041,9 @@ class XpraClient(XpraClientBase, gobject.GObject):
             log.error("failed to start sound sink", exc_info=True)
             return False
 
-    def new_sound_buffer(self, sound_source, data):
+    def new_sound_buffer(self, sound_source, data, metadata):
         assert self.sound_source
-        self.send("sound-data", self.sound_source.codec, Compressed(self.sound_source.codec, data))
+        self.send("sound-data", self.sound_source.codec, Compressed(self.sound_source.codec, data), metadata)
 
     def _process_sound_data(self, packet):
         if not self.speaker_enabled:
@@ -1081,10 +1175,14 @@ class XpraClient(XpraClientBase, gobject.GObject):
             auto_refresh_delay = self.auto_refresh_delay    #we do it
         ClientWindowClass = ClientWindow
         if not self.mmap_enabled and self.opengl_enabled and self.encoding in ("x264", "vpx"):
-            ClientWindowClass = self.GLClientWindowClass
+            #only enable GL for normal windows:
+            window_types = metadata.get("window-type", ())
+            if "_NET_WM_WINDOW_TYPE_NORMAL" in window_types:
+                ClientWindowClass = self.GLClientWindowClass
         pid = metadata.get("pid", -1)
         group_leader = None
-        if pid>0 and not is_gtk3():
+        #set group leader (but avoid ugly "not implemented" warning on win32):
+        if pid>0 and not is_gtk3() and not sys.platform.startswith("win"):
             group_leader = self._pid_to_group_leader.get(pid)
             if not group_leader:
                 if is_gtk3():
@@ -1094,8 +1192,8 @@ class XpraClient(XpraClientBase, gobject.GObject):
                 else:
                     WINDOW_TOPLEVEL = gdk.WINDOW_TOPLEVEL
                     INPUT_ONLY = gdk.INPUT_ONLY
-                group_leader = gdk.Window(None, 1, 1, WINDOW_TOPLEVEL, 0, INPUT_ONLY,
-                                              "group-leader-for-%s" % pid)
+                title = "%s group leader for %s" % (self.session_name or "Xpra", pid)
+                group_leader = gdk.Window(None, 1, 1, WINDOW_TOPLEVEL, 0, INPUT_ONLY, title)
                 self._pid_to_group_leader[pid] = group_leader
                 log("new hidden group leader window %s for pid=%s", group_leader, pid)
             self._group_leader_wids.setdefault(group_leader, []).append(wid)
@@ -1138,6 +1236,8 @@ class XpraClient(XpraClientBase, gobject.GObject):
             try:
                 self._do_draw(packet)
                 time.sleep(0)
+            except KeyboardInterrupt:
+                raise
             except:
                 log.error("error processing draw packet", exc_info=True)
 
@@ -1170,7 +1270,7 @@ class XpraClient(XpraClientBase, gobject.GObject):
             if success:
                 end = time.time()
                 decode_time = int(end*1000*1000-start*1000*1000)
-                self.pixel_counter.append((end, width*height))
+                self.pixel_counter.append((start, end, width*height))
                 if DRAW_DEBUG:
                     dms = "%sms" % (int(decode_time/100)/10.0)
                     log.info("record_decode_time(%s) wid=%s, %s: %sx%s, %s", success, wid, coding, width, height, dms)
@@ -1181,6 +1281,8 @@ class XpraClient(XpraClientBase, gobject.GObject):
             self.send_damage_sequence(wid, packet_sequence, width, height, decode_time)
         try:
             window.draw_region(x, y, width, height, coding, data, rowstride, packet_sequence, options, [record_decode_time])
+        except KeyboardInterrupt:
+            raise
         except:
             log.error("draw error", exc_info=True)
             gobject.idle_add(record_decode_time, False)
@@ -1304,8 +1406,9 @@ class XpraClient(XpraClientBase, gobject.GObject):
         log("set maximum packet size to %s", self._protocol.max_packet_size)
 
     def process_packet(self, proto, packet):
-        packet_type = str(packet[0])
-        if packet_type.startswith("clipboard-"):
+        packet_type = packet[0]
+        self.check_server_echo(0)
+        if type(packet_type) in (unicode, str) and packet_type.startswith("clipboard-"):
             if self.clipboard_enabled:
                 gobject.idle_add(self._client_extras.process_clipboard_packet, packet)
         else:
