@@ -6,6 +6,8 @@
 
 import os
 import struct
+import re
+import binascii
 
 from wimpiggy.gobject_compat import import_gobject, import_gtk, import_gdk, is_gtk3
 gobject = import_gobject()
@@ -16,7 +18,6 @@ if is_gtk3():
 else:
     PROPERTY_CHANGE_MASK = gdk.PROPERTY_CHANGE_MASK
 
-
 from wimpiggy.util import n_arg_signal
 from wimpiggy.log import Logger
 log = Logger()
@@ -24,41 +25,97 @@ log = Logger()
 from xpra.nested_main import NestedMainLoop
 from xpra.protocol import zlib_compress
 
-#debug = log.info
-debug = log.debug
+if os.environ.get("XPRA_CLIPBOARD_DEBUG", "0")=="1":
+    debug = log.info
+else:
+    debug = log.debug
+
+MAX_CLIPBOARD_PACKET_SIZE = 256*1024
+
+CLIPBOARDS = os.environ.get("XPRA_CLIPBOARDS", "CLIPBOARD,PRIMARY,SECONDARY").split(",")
+CLIPBOARDS = [x.upper().strip() for x in CLIPBOARDS]
+
+_discard_target_strs_ = ("^SAVE_TARGETS$",
+        "^COMPOUND_TEXT$",
+        "^NeXT ",
+        "^com\.apple\."
+        "^CorePasteboardFlavorType "
+        "^dyn\.")
+DISCARD_TARGETS = [re.compile(x) for x in _discard_target_strs_]
+
 
 class ClipboardProtocolHelperBase(object):
-    def __init__(self, send_packet_cb, clipboards=["CLIPBOARD"]):
+    def __init__(self, send_packet_cb, progress_cb=None, clipboards=CLIPBOARDS, filter_res=None):
         self.send = send_packet_cb
-        self.max_clipboard_packet_size = 32*1024 - 1024
+        self.progress_cb = progress_cb
+        self.max_clipboard_packet_size = MAX_CLIPBOARD_PACKET_SIZE
+        self.filter_res = []
+        if filter_res:
+            for x in filter_res:
+                try:
+                    self.filter_res.append(re.compile(x))
+                except:
+                    log.error("invalid regular expression '%s' in clipboard filter")
+        self._clipboard_request_counter = 0
+        self._clipboard_outstanding_requests = {}
+        self._want_targets = False
+        self.init_packet_handlers()
+        self.init_proxies(clipboards)
+
+    def set_greedy_client(self, greedy):
+        for proxy in self._clipboard_proxies.values():
+            proxy.set_greedy_client(greedy)
+
+    def set_want_targets_client(self, want_targets):
+        debug("set_want_targets_client(%s)", want_targets)
+        self._want_targets = want_targets
+
+    def init_packet_handlers(self):
+        self._packet_handlers = {
+            "clipboard-token":              self._process_clipboard_token,
+            "clipboard-request":            self._process_clipboard_request,
+            "clipboard-contents":           self._process_clipboard_contents,
+            "clipboard-contents-none":      self._process_clipboard_contents_none,
+            "clipboard-pending-requests":   self._process_clipboard_pending_requests,
+            }
+
+    def make_proxy(self, clipboard):
+        return ClipboardProxy(clipboard)
+
+    def init_proxies(self, clipboards):
         self._clipboard_proxies = {}
         for clipboard in clipboards:
-            proxy = ClipboardProxy(clipboard)
+            proxy = self.make_proxy(clipboard)
             proxy.connect("send-clipboard-token", self._send_clipboard_token_handler)
             proxy.connect("get-clipboard-from-remote", self._get_clipboard_from_remote_handler)
             proxy.show()
             self._clipboard_proxies[clipboard] = proxy
-        self._clipboard_request_counter = 0
-        self._clipboard_outstanding_requests = {}
+        debug("%s.init_proxies : %s", type(self), self._clipboard_proxies)
 
     def local_to_remote(self, selection):
+        #overriden in some subclasses (see: translated_clipboard)
         return  selection
     def remote_to_local(self, selection):
+        #overriden in some subclasses (see: translated_clipboard)
         return  selection
 
     # Used by the client during startup:
     def send_all_tokens(self):
-        for selection in self._clipboard_proxies:
-            name = self.local_to_remote(selection)
-            debug("send_all_tokens selection=%s, exported as=%s", selection, name)
-            self.send("clipboard-token", name)
+        for selection, proxy in self._clipboard_proxies.items():
+            self._send_clipboard_token_handler(proxy, selection)
 
     def _process_clipboard_token(self, packet):
         selection = packet[1]
+        targets = None
+        if len(packet)>=3:
+            targets = packet[2]
         name = self.remote_to_local(selection)
-        debug("process clipboard token selection=%s, local clipboard name=%s", selection, name)
-        if name in self._clipboard_proxies:
-            self._clipboard_proxies[name].got_token()
+        proxy = self._clipboard_proxies.get(name)
+        debug("process clipboard token selection=%s, local clipboard name=%s, proxy=%s", selection, name, proxy)
+        if proxy:
+            proxy.got_token(targets)
+        else:
+            debug("ignoring token for clipboard proxy name '%s' (no proxy)", name)
 
     def _get_clipboard_from_remote_handler(self, proxy, selection, target):
         request_id = self._clipboard_request_counter
@@ -66,23 +123,36 @@ class ClipboardProtocolHelperBase(object):
         debug("get clipboard from remote handler id=%s", request_id)
         loop = NestedMainLoop()
         self._clipboard_outstanding_requests[request_id] = loop
+        if self.progress_cb:
+            self.progress_cb(len(self._clipboard_outstanding_requests), None)
         self.send("clipboard-request", request_id, self.local_to_remote(selection), target)
         result = loop.main(1 * 1000, 2 * 1000)
         debug("get clipboard from remote result(%s)=%s", request_id, result)
         del self._clipboard_outstanding_requests[request_id]
+        if self.progress_cb:
+            self.progress_cb(len(self._clipboard_outstanding_requests), None)
         return result
 
     def _clipboard_got_contents(self, request_id, dtype, dformat, data):
-        debug("got clipboard contents(%s)=%s (type=%s, format=%s)", request_id, len(data or []), dtype, dformat)
-        if request_id in self._clipboard_outstanding_requests:
-            loop = self._clipboard_outstanding_requests[request_id]
-            loop.done({"type": dtype, "format": dformat, "data": data})
-        else:
+        loop = self._clipboard_outstanding_requests.get(request_id)
+        debug("got clipboard contents for id=%s len=%s, loop=%s (type=%s, format=%s)",
+              request_id, len(data or []), loop, dtype, dformat)
+        if loop is None:
             debug("got unexpected response to clipboard request %s", request_id)
+            return
+        loop.done({"type": dtype, "format": dformat, "data": data})
 
     def _send_clipboard_token_handler(self, proxy, selection):
         debug("send clipboard token: %s", selection)
-        self.send("clipboard-token", self.local_to_remote(selection))
+        rsel = self.local_to_remote(selection)
+        if self._want_targets:
+            #send the token with the target once we get them:
+            def got_targets(dtype, dformat, targets):
+                debug("sending clipboard token for %s with targets=%s", selection, targets)
+                self.send("clipboard-token", rsel, targets)
+            proxy.get_contents("TARGETS", got_targets)
+            return
+        self.send("clipboard-token", rsel)
 
     def _munge_raw_selection_to_wire(self, target, dtype, dformat, data):
         # Some types just cannot be marshalled:
@@ -90,8 +160,26 @@ class ClipboardProtocolHelperBase(object):
                     "PIXEL", "COLORMAP"):
             debug("skipping clipboard data of type: %s, format=%s, len(data)=%s", dtype, dformat, len(data))
             return None, None
+        if target=="TARGETS" and dtype=="ATOM":
+            #targets is special cased here
+            #because we get the values in wire format already (not atoms)
+            #thanks to the request_targets() function (required on win32)
+            return "atoms", self._filter_targets(data)
         return self._do_munge_raw_selection_to_wire(target, dtype, dformat, data)
 
+    def _filter_targets(self, targets):
+        remove = []
+        for target in targets:
+            for x in DISCARD_TARGETS:
+                if x.match(target):
+                    remove.append(target)
+                    break
+        f = list(targets)
+        for t in remove:
+            f.remove(t)
+        debug("_filter_targets(%s)=%s", targets, f)
+        return f
+
     def _do_munge_raw_selection_to_wire(self, target, dtype, dformat, data):
         """ this method is overriden in xclipboard to parse X11 atoms """
         # Other types need special handling, and all types need to be
@@ -116,6 +204,10 @@ class ClipboardProtocolHelperBase(object):
             ints = struct.unpack(binfmt, data)
             return "integers", ints
         elif dformat == 8:
+            for x in self.filter_res:
+                if x.match(data):
+                    log.warn("clipboard buffer contains blacklisted pattern '%s' and has been dropped!", x.pattern)
+                    return None, None
             return "bytes", data
         else:
             log.error("unhandled format %s for clipboard data type %s" % (dformat, dtype))
@@ -141,35 +233,38 @@ class ClipboardProtocolHelperBase(object):
             raise Exception("unhanled encoding: %s" % encoding)
 
     def _process_clipboard_request(self, packet):
+        def no_contents():
+            self.send("clipboard-contents-none", request_id, selection)
         request_id, selection, target = packet[1:4]
         name = self.remote_to_local(selection)
         debug("process clipboard request, request_id=%s, selection=%s, local name=%s, target=%s", request_id, selection, name, target)
-        if name in self._clipboard_proxies:
-            proxy = self._clipboard_proxies[name]
-            def got_contents(dtype, dformat, data):
-                debug("got_contents(%s, %s, %s:%s) data=%s, str(data)=%s", dtype, dformat, type(data), len(data or ""), list(bytearray(data or "")[:200]), str(data)[:200])
-                def no_contents():
-                    self.send("clipboard-contents-none", request_id, selection)
-                if dtype is None or data is None:
+        proxy = self._clipboard_proxies.get(name)
+        if proxy is None:
+            #err, we were asked about a clipboard we don't handle..
+            no_contents()
+            return
+        def got_contents(dtype, dformat, data):
+            debug("got_contents(%s, %s, %s:%s) data=0x%s..",
+                  dtype, dformat, type(data), len(data or ""), binascii.hexlify(str(data)[:200]))
+            if dtype is None or data is None:
+                no_contents()
+                return
+            munged = self._munge_raw_selection_to_wire(target, dtype, dformat, data)
+            wire_encoding, wire_data = munged
+            debug("clipboard raw -> wire: %r -> %r", (dtype, dformat, data), munged)
+            if wire_encoding is None:
+                no_contents()
+                return
+            if len(wire_data)>256:
+                wire_data = zlib_compress("clipboard: %s / %s" % (dtype, dformat), wire_data)
+                if len(wire_data)>self.max_clipboard_packet_size:
+                    log.warn("even compressed, clipboard contents are too big and have not been sent:"
+                             " %s compressed bytes dropped (maximum is %s)", len(wire_data), self.max_clipboard_packet_size)
                     no_contents()
                     return
-                munged = self._munge_raw_selection_to_wire(target, dtype, dformat, data)
-                wire_encoding, wire_data = munged
-                debug("clipboard raw -> wire: %r -> %r", (dtype, dformat, data), munged)
-                if wire_encoding is None:
-                    no_contents()
-                    return
-                if len(wire_data)>256:
-                    wire_data = zlib_compress("clipboard: %s / %s" % (dtype, dformat), wire_data)
-                    if len(wire_data)>self.max_clipboard_packet_size:
-                        log.warn("even compressed, clipboard contents are too big and have not been sent: %s compressed bytes dropped" % len(wire_data))
-                        no_contents()
-                        return
-                self.send("clipboard-contents", request_id, selection,
-                           dtype, dformat, wire_encoding, wire_data)
-            proxy.get_contents(target, got_contents)
-        else:
-            self.send("clipboard-contents-none", request_id, selection)
+            self.send("clipboard-contents", request_id, selection,
+                       dtype, dformat, wire_encoding, wire_data)
+        proxy.get_contents(target, got_contents)
 
     def _process_clipboard_contents(self, packet):
         request_id, selection, dtype, dformat, wire_encoding, wire_data = packet[1:8]
@@ -183,17 +278,15 @@ class ClipboardProtocolHelperBase(object):
         request_id = packet[1]
         self._clipboard_got_contents(request_id, None, None, None)
 
-    _packet_handlers = {
-        "clipboard-token": _process_clipboard_token,
-        "clipboard-request": _process_clipboard_request,
-        "clipboard-contents": _process_clipboard_contents,
-        "clipboard-contents-none": _process_clipboard_contents_none,
-        }
+    def _process_clipboard_pending_requests(self, packet):
+        pending = packet[1]
+        if self.progress_cb:
+            self.progress_cb(None, pending)
 
     def process_clipboard_packet(self, packet):
         packet_type = packet[0]
         debug("process clipboard packet type=%s", packet_type)
-        self._packet_handlers[packet_type](self, packet)
+        self._packet_handlers[packet_type](packet)
 
 
 class DefaultClipboardProtocolHelper(ClipboardProtocolHelperBase):
@@ -202,8 +295,8 @@ class DefaultClipboardProtocolHelper(ClipboardProtocolHelperBase):
         But without gdk atom support, see gdk_clipboard for a better one!
     """
 
-    def __init__(self, send_packet_cb):
-        ClipboardProtocolHelperBase.__init__(self, send_packet_cb, ["CLIPBOARD", "PRIMARY", "SECONDARY"])
+    def __init__(self, send_packet_cb, progress_cb=None):
+        ClipboardProtocolHelperBase.__init__(self, send_packet_cb, progress_cb, ["CLIPBOARD", "PRIMARY", "SECONDARY"])
 
 
 class ClipboardProxy(gtk.Invisible):
@@ -223,6 +316,24 @@ class ClipboardProxy(gtk.Invisible):
         self._selection = selection
         self._clipboard = gtk.Clipboard(selection=selection)
         self._have_token = False
+        #clients that need a new token for every owner-change: (ie: win32)
+        self._greedy_client = False
+        #semaphore to block the sending of the token when we change the owner ourselves:
+        self._block_owner_change = False
+        self._clipboard.connect("owner-change", self.do_owner_changed)
+
+    def set_greedy_client(self, greedy):
+        debug("%s.set_greedy_client(%s)", self, greedy)
+        self._greedy_client = greedy
+
+    def __str__(self):
+        return  "ClipboardProxy(%s)" % self._selection
+
+    def do_owner_changed(self, *args):
+        debug("do_owner_changed(%s) greedy_client=%s, block_owner_change=%s", args, self._greedy_client, self._block_owner_change)
+        if self._greedy_client and not self._block_owner_change:
+            self._have_token = False
+            self.emit("send-clipboard-token", self._selection)
 
     def do_selection_request_event(self, event):
         debug("do_selection_request_event(%s)", event)
@@ -263,20 +374,16 @@ class ClipboardProxy(gtk.Invisible):
         elif target == "MULTIPLE":
             try:
                 from wimpiggy.prop import prop_get
-                from wimpiggy.error import trap
             except ImportError:
                 debug("MULTIPLE for property '%s' not handled due to missing wimpiggy bindings", event.property)
                 gtk.Invisible.do_selection_request_event(self, event)
                 return
-            def get_targets(targets):
-                atoms = prop_get(event.window, event.property, ["multiple-conversion"])
-                debug("MULTIPLE clipboard atoms: %r", atoms)
-                targets += atoms[::2]
-            targets = []
-            trap.swallow(get_targets, targets)
-            debug("MULTIPLE clipboard targets: %r", targets)
-            for target in targets:
-                self.selection_add_target(self._selection, target, 0)
+            atoms = prop_get(event.window, event.property, ["multiple-conversion"])
+            debug("MULTIPLE clipboard atoms: %r", atoms)
+            if atoms:
+                targets = atoms[::2]
+                for t in targets:
+                    self.selection_add_target(self._selection, t, 0)
         else:
             debug("target for %s: %r", self._selection, target)
             self.selection_add_target(self._selection, target, 0)
@@ -289,34 +396,41 @@ class ClipboardProxy(gtk.Invisible):
         # Either call selection_data.set() or don't, and then return.
         # In practice, send a call across the wire, then block in a recursive
         # main loop.
-        debug("do_selection_get(%s, %s, %s)", selection_data, info, time)
+        debug("do_selection_get(%s, %s, %s) selection=%s", selection_data, info, time, selection_data.selection)
         assert self._selection == str(selection_data.selection)
         target = str(selection_data.target)
         result = self.emit("get-clipboard-from-remote", self._selection, target)
-        if result is not None and result["type"] is not None:
-            data = result["data"]
-            dformat = result["format"]
-            dtype = result["type"]
-            debug("do_selection_get(%s,%s,%s) calling selection_data.set(%s, %s, %s:%s)", selection_data, info, time, dtype, dformat, type(data), len(data or ""))
-            selection_data.set(dtype, dformat, data)
-        else:
-            debug("remote selection fetch timed out")
+        if result is None or result["type"] is None:
+            debug("remote selection fetch timed out or empty")
+            return
+        data = result["data"]
+        dformat = result["format"]
+        dtype = result["type"]
+        debug("do_selection_get(%s,%s,%s) calling selection_data.set(%s, %s, %s:%s)",
+              selection_data, info, time, dtype, dformat, type(data), len(data or ""))
+        selection_data.set(dtype, dformat, data)
 
     def do_selection_clear_event(self, event):
         # Someone else on our side has the selection
         debug("do_selection_clear_event(%s) selection=%s", event, self._selection)
+        #if greedy_client is set, do_owner_changed will fire the token
+        #so don't bother sending it now (same if we don't have it)
+        send = ((not self._greedy_client or self._block_owner_change) or not self._have_token)
         self._have_token = False
 
         # Emit a signal -> send a note to the other side saying "hey its
         # ours now"
         # Send off the anti-token.
-        self.emit("send-clipboard-token", self._selection)
+        if send:
+            self.emit("send-clipboard-token", self._selection)
         gtk.Invisible.do_selection_clear_event(self, event)
 
-    def got_token(self):
+    def got_token(self, targets):
         # We got the anti-token.
-        debug("got token, selection=%s", self._selection)
+        debug("got token, selection=%s, targets=%s", self._selection, targets)
         self._have_token = True
+        if self._greedy_client:
+            self._block_owner_change = True
         if not self.selection_owner_set(self._selection):
             # I don't know how this can actually fail, given that we pass
             # CurrentTime, but just in case:
@@ -324,25 +438,37 @@ class ClipboardProxy(gtk.Invisible):
                      % (self._selection,)
                      + "will not be able to pass local apps "
                      + "contents of remote clipboard")
+        if self._block_owner_change:
+            #re-enable the flag via idle_add so events like do_owner_changed
+            #get a chance to run first.
+            def remove_block(*args):
+                self._block_owner_change = False
+            gobject.idle_add(remove_block)
 
     # This function is called by the xpra core when the peer has requested the
     # contents of this clipboard:
     def get_contents(self, target, cb):
-        debug("get_contents(%s,%s)", target, cb)
+        debug("get_contents(%s,%s) selection=%s", target, cb, self._selection)
         if self._have_token:
             log.warn("Our peer requested the contents of the clipboard, but "
                      + "*I* thought *they* had it... weird.")
             cb(None, None, None)
-        def unpack(clipboard, selection_data, data):
-            debug("unpack(%s, %s, %s:%s)", clipboard, selection_data, type(data), len(data or ""))
+            return
+        if target=="TARGETS":
+            #handle TARGETS using "request_targets"
+            def got_targets(c, targets, *args):
+                debug("got_targets(%s, %s, %s)", c, targets, args)
+                cb("ATOM", 32, targets)
+            self._clipboard.request_targets(got_targets)
+            return
+        def unpack(clipboard, selection_data, user_data):
+            debug("unpack: %s", type(selection_data))
             if selection_data is None:
                 cb(None, None, None)
-            else:
-                debug("unpack(..) type=%s, format=%s, data=%s:%s", selection_data.type, selection_data.format,
-                            type(selection_data.data), len(selection_data.data or ""))
-                cb(str(selection_data.type),
-                   selection_data.format,
-                   selection_data.data)
+                return
+            debug("unpack(..) type=%s, format=%s, data=%s:%s", selection_data.type, selection_data.format,
+                        type(selection_data.data), len(selection_data.data or ""))
+            cb(str(selection_data.type), selection_data.format, selection_data.data)
         self._clipboard.request_contents(target, unpack)
 
 gobject.type_register(ClipboardProxy)
