@@ -17,7 +17,10 @@ from zlib import compress, decompress, decompressobj
 import struct
 import os
 import threading
+import errno
+import binascii
 
+USE_ALIASES = os.environ.get("XPRA_USE_ALIASES", "1")=="1"
 PACKET_JOIN_SIZE = int(os.environ.get("XPRA_PACKET_JOIN_SIZE", 16384))
 
 if sys.version_info[:2]>=(2,5):
@@ -31,24 +34,26 @@ try:
     from queue import Queue     #@UnresolvedImport @UnusedImport (python3)
 except:
     from Queue import Queue     #@Reimport
-from threading import RLock
+from threading import Lock
 
 from wimpiggy.log import Logger
 log = Logger()
 
 from xpra.daemon_thread import make_daemon_thread
-from xpra.bytestreams import untilConcludes
 from xpra.bencode import bencode, bdecode
-rencode_dumps, rencode_loads = None, None
+rencode_dumps, rencode_loads, rencode_version = None, None, None
 try:
     try:
         from xpra.rencode import dumps as rencode_dumps  #@UnresolvedImport
         from xpra.rencode import loads as rencode_loads  #@UnresolvedImport
+        from xpra.rencode import __version__
+        rencode_version = __version__
+        log("found rencode version %s", rencode_version)
     except ImportError:
-        pass
+        log.error("rencode load error", exc_info=True)
 except Exception, e:
     log.error("xpra.rencode is missing: %s", e)
-has_rencode = rencode_dumps is not None and rencode_loads is not None
+has_rencode = rencode_dumps is not None and rencode_loads is not None and rencode_version is not None
 use_rencode = has_rencode and not os.environ.get("XPRA_USE_BENCODER", "0")=="1"
 
 #stupid python version breakage:
@@ -74,6 +79,8 @@ class Compressed(object):
         self.data = data
     def __len__(self):
         return len(self.data)
+    def __str__(self):
+        return  "Compressed(%s: %s bytes)" % (self.datatype, len(self.data))
 
 class ZLibCompressed(object):
     def __init__(self, datatype, data, level):
@@ -82,6 +89,9 @@ class ZLibCompressed(object):
         self.level = level
     def __len__(self):
         return len(self.data)
+    def __str__(self):
+        return  "ZLibCompressed(%s: %s bytes)" % (self.datatype, len(self.data))
+
 
 def zlib_compress(datatype, data, level=5):
     cdata = zcompress(data, level)
@@ -114,6 +124,7 @@ class Protocol(object):
         #initial value which may get increased by client/server after handshake:
         self.max_packet_size = 32*1024
         self.large_packets = ["hello"]
+        self.aliases = {}
         self.chunked_compression = True
         self._closed = False
         self._encoder = self.bencode
@@ -125,7 +136,7 @@ class Protocol(object):
         self.cipher_out = None
         self.cipher_out_name = None
         self.cipher_out_block_size = 0
-        self._write_lock = RLock()
+        self._write_lock = Lock()
         self._write_thread = make_daemon_thread(self._write_thread_loop, "write")
         self._read_thread = make_daemon_thread(self._read_thread_loop, "read")
         self._read_parser_thread = make_daemon_thread(self._read_parse_thread_loop, "parse")
@@ -186,6 +197,15 @@ class Protocol(object):
                 self._write_format_thread.start()
         gobject.idle_add(do_start)
 
+    def send_now(self, packet):
+        log("send_now(%s ...)", packet[0])
+        assert self._get_packet_cb==None, "cannot use send_now when a packet source exists!"
+        def packet_cb():
+            self._get_packet_cb = None
+            return (packet, )
+        self._get_packet_cb = packet_cb
+        self.source_has_more()
+
     def source_has_more(self):
         self._source_has_more.set()
 
@@ -202,50 +222,54 @@ class Protocol(object):
             self._source_has_more.set()
         if packet is None:
             return
-        packets, proto_flags = self.encode(packet)
+        chunks, proto_flags = self.encode(packet)
         try:
             self._write_lock.acquire()
-            counter = 0
-            items = []
-            for index,level,data in packets:
-                payload_size = len(data)
-                actual_size = payload_size
-                scb, ecb = None, None
-                #fire the start_send_callback just before the first packet is processed:
-                if counter==0:
-                    scb = start_send_cb
-                #fire the end_send callback when the last packet (index==0) makes it out:
-                if index==0:
-                    ecb = end_send_cb
-                if self.cipher_out:
-                    proto_flags |= Protocol.FLAGS_CIPHER
-                    #note: since we are padding: l!=len(data)
-                    padding = (self.cipher_out_block_size - len(data) % self.cipher_out_block_size) * " "
-                    if len(padding)==0:
-                        padded = data
-                    else:
-                        padded = data+padding
-                    actual_size = payload_size + len(padding)
-                    assert len(padded)==actual_size
-                    data = self.cipher_out.encrypt(padded)
-                    assert len(data)==actual_size
-                    log("sending %s bytes encrypted with %s padding", payload_size, len(padding))
-                if actual_size<PACKET_JOIN_SIZE:
-                    #'p' + protocol-flags + compression_level + packet_index + data_size
-                    if type(data)==unicode:
-                        data = str(data)
-                    header_and_data = struct.pack('!BBBBL%ss' % actual_size, ord("P"), proto_flags, level, index, payload_size, data)
-                    items.append((header_and_data, scb, ecb))
-                else:
-                    header = struct.pack('!BBBBL', ord("P"), proto_flags, level, index, payload_size)
-                    items.append((header, scb, None))
-                    items.append((data, None, ecb))
-                counter += 1
-            self._write_queue.put(items)
+            self._add_chunks_to_queue(chunks, proto_flags, start_send_cb, end_send_cb)
         finally:
-            self.output_packetcount += 1
             self._write_lock.release()
 
+    def _add_chunks_to_queue(self, chunks, proto_flags, start_send_cb=None, end_send_cb=None):
+        """ the write_lock must be held when calling this function """
+        counter = 0
+        items = []
+        for index,level,data in chunks:
+            scb, ecb = None, None
+            #fire the start_send_callback just before the first packet is processed:
+            if counter==0:
+                scb = start_send_cb
+            #fire the end_send callback when the last packet (index==0) makes it out:
+            if index==0:
+                ecb = end_send_cb
+            payload_size = len(data)
+            actual_size = payload_size
+            if self.cipher_out:
+                proto_flags |= Protocol.FLAGS_CIPHER
+                #note: since we are padding: l!=len(data)
+                padding = (self.cipher_out_block_size - len(data) % self.cipher_out_block_size) * " "
+                if len(padding)==0:
+                    padded = data
+                else:
+                    padded = data+padding
+                actual_size = payload_size + len(padding)
+                assert len(padded)==actual_size
+                data = self.cipher_out.encrypt(padded)
+                assert len(data)==actual_size
+                log("sending %s bytes encrypted with %s padding", payload_size, len(padding))
+            if actual_size<PACKET_JOIN_SIZE:
+                #'p' + protocol-flags + compression_level + packet_index + data_size
+                if type(data)==unicode:
+                    data = str(data)
+                header_and_data = struct.pack('!BBBBL%ss' % actual_size, ord("P"), proto_flags, level, index, payload_size, data)
+                items.append((header_and_data, scb, ecb))
+            else:
+                header = struct.pack('!BBBBL', ord("P"), proto_flags, level, index, payload_size)
+                items.append((header, scb, None))
+                items.append((data, None, ecb))
+            counter += 1
+        self._write_queue.put(items)
+        self.output_packetcount += 1
+
     def verify_packet(self, packet):
         """ look for None values which may have caused the packet to fail encoding """
         if type(packet)!=list:
@@ -326,14 +350,16 @@ class Protocol(object):
             elif ti!=str:
                 log.info("unexpected data type in %s packet: %s", packet[0], ti)
         #now the main packet (or what is left of it):
+        packet_type = packet[0]
+        if USE_ALIASES and self.aliases and packet_type in self.aliases:
+            #replace the packet type with the alias:
+            packet[0] = self.aliases[packet_type]
         try:
             main_packet, proto_version = self._encoder(packet)
         except (KeyError, TypeError), e:
             if self._closed:
                 return [], 0
-            log.error("failed to encode packet: %s", packet)
-            import traceback
-            traceback.print_exc()
+            log.error("failed to encode packet: %s", packet, exc_info=True)
             self.verify_packet(packet)
             raise e
         if len(main_packet)>=1024 and packet_in[0] not in self.large_packets:
@@ -350,57 +376,64 @@ class Protocol(object):
         #this may be used next time encode() is called
         self._compression_level = level
 
-    def _write_thread_loop(self):
+    def _io_thread_loop(self, name, callback):
         try:
             while not self._closed:
-                items = self._write_queue.get()
-                # Used to signal that we should exit:
-                if items is None:
-                    log("write thread: empty marker, exiting")
-                    break
-                for buf, start_cb, end_cb in items:
-                    if start_cb:
-                        try:
-                            start_cb(self._conn.output_bytecount)
-                        except:
-                            log.error("error on %s", start_cb, exc_info=True)
-                    while buf and not self._closed:
-                        written = untilConcludes(self._conn.write, buf)
-                        if written:
-                            buf = buf[written:]
-                            self.output_raw_packetcount += 1
-                    if end_cb:
-                        try:
-                            end_cb(self._conn.output_bytecount)
-                        except:
-                            if not self._closed:
-                                log.error("error on %s", end_cb, exc_info=True)
-            log("write thread: ended")
+                callback()
+        except KeyboardInterrupt, e:
+            raise e
         except (OSError, IOError, socket_error), e:
             if not self._closed:
-                self._call_connection_lost("Error writing to connection: %s" % e)
+                if e.args[0] in (errno.ECONNRESET, errno.EPIPE):
+                    log.error("%s connection reset for %s", name, self._conn)
+                    self._call_connection_lost("%s connection reset: %s" % (name, e))
+                else:
+                    log.error("%s error for %s", name, self._conn, exc_info=True)
+                    self._call_connection_lost("%s error on connection: %s" % (name, e))
         except Exception, e:
             #can happen during close(), in which case we just ignore:
             if not self._closed:
+                log.error("%s error on %s", name, self._conn, exc_info=True)
                 self.close()
-                raise e
+
+    def _write_thread_loop(self):
+        self._io_thread_loop("write", self._write)
+    def _write(self):
+        items = self._write_queue.get()
+        # Used to signal that we should exit:
+        if items is None:
+            log("write thread: empty marker, exiting")
+            self.close()
+            return
+        for buf, start_cb, end_cb in items:
+            if start_cb:
+                try:
+                    start_cb(self._conn.output_bytecount)
+                except:
+                    log.error("error on %s", start_cb, exc_info=True)
+            while buf and not self._closed:
+                written = self._conn.write(buf)
+                if written:
+                    buf = buf[written:]
+                    self.output_raw_packetcount += 1
+            if end_cb:
+                try:
+                    end_cb(self._conn.output_bytecount)
+                except:
+                    if not self._closed:
+                        log.error("error on %s", end_cb, exc_info=True)
 
     def _read_thread_loop(self):
-        try:
-            while not self._closed:
-                buf = untilConcludes(self._conn.read, 8192)
-                #log("read thread: got data of size %s: %s", len(buf), repr_ellipsized(buf))
-                self._read_queue.put(buf)
-                if not buf:
-                    log("read thread: eof")
-                    break
-                self.input_raw_packetcount += 1
-        except (ValueError, OSError, IOError, socket_error), e:
-            if not self._closed:
-                self._call_connection_lost("Error reading from connection: %s" % e)
-        except Exception, e:
-            if not self._closed:
-                self._call_connection_lost("Error reading from connection: %s" % e)
+        self._io_thread_loop("read", self._read)
+    def _read(self):
+        buf = self._conn.read(8192)
+        #log("read thread: got data of size %s: %s", len(buf), repr_ellipsized(buf))
+        self._read_queue.put(buf)
+        if not buf:
+            log("read thread: eof")
+            self.close()
+            return
+        self.input_raw_packetcount += 1
 
     def _call_connection_lost(self, message="", exc_info=False):
         log("will call connection lost: %s", message)
@@ -451,14 +484,19 @@ class Protocol(object):
                     break
                 if payload_size<0:
                     if read_buffer[0] not in ("P", ord("P")):
-                        return self._call_connection_lost("invalid packet header byte: ('%s...'), not an xpra client?" % read_buffer[:32])
+                        err = "invalid packet header byte: '%s', not an xpra client?" % hex(ord(read_buffer[0]))
+                        if len(read_buffer)>1:
+                            err += " read buffer=0x%s" % binascii.hexlify(read_buffer[:40])
+                            if len(read_buffer)>40:
+                                err += "..."
+                        return self._call_connection_lost(err)
                     if bl<8:
                         break   #packet still too small
                     #packet format: struct.pack('cBBBL', ...) - 8 bytes
                     try:
                         _, protocol_flags, compression_level, packet_index, data_size = unpack_header(read_buffer[:8])
                     except Exception, e:
-                        raise Exception("failed to parse packet header: %s" % list(read_buffer[:8]), e)
+                        raise Exception("failed to parse packet header: 0x%s" % binascii.hexlify(read_buffer[:8]), e)
                     read_buffer = read_buffer[8:]
                     bl = len(read_buffer)
                     if protocol_flags & Protocol.FLAGS_CIPHER:
@@ -475,10 +513,10 @@ class Protocol(object):
                     #this packet is seemingly too big, but check again from the main UI thread
                     #this gives 'set_max_packet_size' a chance to run from "hello"
                     def check_packet_size(size_to_check, packet_header):
-                        log("check_packet_size(%s, %s) limit is %s", size_to_check, packet_header, self.max_packet_size)
+                        log("check_packet_size(%s, 0x%s) limit is %s", size_to_check, binascii.hexlify(packet_header), self.max_packet_size)
                         if size_to_check>self.max_packet_size:
-                            return self._call_connection_lost("invalid packet: size requested is %s (maximum allowed is %s - packet header: '%s'), dropping this connection!" %
-                                                              (size_to_check, self.max_packet_size, packet_header))
+                            return self._call_connection_lost("invalid packet: size requested is %s (maximum allowed is %s - packet header: 0x%s), dropping this connection!" %
+                                                              (size_to_check, self.max_packet_size, binascii.hexlify(packet_header)))
                     gobject.timeout_add(1000, check_packet_size, payload_size, read_buffer[:32])
 
                 if bl<payload_size:
@@ -498,7 +536,13 @@ class Protocol(object):
                     log("received %s encrypted bytes with %s padding", payload_size, len(padding))
                     data = self.cipher_in.decrypt(raw_string)
                     if padding:
-                        assert data.endswith(padding), "decryption failed: string does not end with '%s': %s (%s) -> %s (%s)" % (padding, list(bytearray(raw_string)), type(raw_string), list(bytearray(data)), type(data))
+                        def debug_str():
+                            try:
+                                return list(bytearray(raw_string))
+                            except:
+                                return list(str(raw_string))
+                        assert data.endswith(padding), "decryption failed: string does not end with '%s': %s (%s) -> %s (%s)" % \
+                            (padding, debug_str(raw_string), type(raw_string), debug_str(data), type(data))
                         data = data[:-len(padding)]
                 #uncompress if needed:
                 if compression_level>0:
@@ -531,9 +575,7 @@ class Protocol(object):
                         packet, l = bdecode(data)
                         assert l==len(data)
                 except ValueError, e:
-                    import traceback
-                    traceback.print_exc()
-                    log.error("value error reading packet: %s", e)
+                    log.error("value error reading packet: %s", e, exc_info=True)
                     if self._closed:
                         return
                     def gibberish(buf):
@@ -563,35 +605,53 @@ class Protocol(object):
                     log.warn("Unhandled error while processing a '%s' packet from peer", packet[0], exc_info=True)
 
     def flush_then_close(self, last_packet):
-        try:
-            self._write_lock.acquire()
-            #try to wait for the queue to empty with the lock held
-            try:
-                import time
-                i = 0
-                while not self._closed and not self._write_queue.empty() and i<5:
-                    time.sleep(0.1)
-                    i += 1
-            except:
-                pass
-            #and send our last_packet to it:
-            self._add_packet_to_queue(last_packet)
-        finally:
-            self._write_lock.release()
-        self.terminate_io_threads()
-        #wait for last_packet to be sent:
-        def wait_for_end_of_write(timeout=15):
-            log("wait_for_end_of_write(%s) closed=%s, size=%s", timeout, self._closed, self._write_queue.qsize())
-            if self._closed:
-                """ client has disconnected """
-                return
-            if self._write_queue.empty() or timeout<=0:
-                """ threads have terminated or we timedout """
-                self.close()
+        """ Note: this is best effort only
+            the packet may not get sent.
+
+            We try to get the write lock,
+            we try to wait for the write queue to flush
+            we queue our last packet,
+            we wait again for the queue to flush,
+            then no matter what, we close the connection and stop the threads.
+        """
+        def wait_for_queue(timeout=10):
+            #IMPORTANT: if we are here, we have the write lock held!
+            if not self._write_queue.empty():
+                #write queue still has stuff in it..
+                if timeout<=0:
+                    log("flush_then_close: queue still busy, closing without sending the last packet")
+                    self._write_lock.release()
+                    self.close()
+                else:
+                    log("flush_then_close: still waiting for queue to flush")
+                    gobject.timeout_add(100, wait_for_queue, timeout-1)
             else:
-                """ check again soon: """
-                gobject.timeout_add(200, wait_for_end_of_write, timeout-1)
-        wait_for_end_of_write()
+                log("flush_then_close: queue is now empty, sending the last packet and closing")
+                chunks, proto_flags = self.encode(last_packet)
+                def close_cb(*args):
+                    self.close()
+                self._add_chunks_to_queue(chunks, proto_flags, start_send_cb=None, end_send_cb=close_cb)
+                self._write_lock.release()
+                gobject.timeout_add(5*1000, self.close)
+
+        def wait_for_write_lock(timeout=100):
+            if not self._write_lock.acquire(False):
+                if timeout<=0:
+                    log("flush_then_close: timeout waiting for the write lock")
+                    self.close()
+                else:
+                    log("flush_then_close: write lock is busy, will retry %s more times", timeout)
+                    gobject.timeout_add(10, wait_for_write_lock, timeout-1)
+            else:
+                log("flush_then_close: acquired the write lock")
+                #we have the write lock - we MUST free it!
+                wait_for_queue()
+        #normal codepath:
+        # -> wait_for_write_lock
+        # -> wait_for_queue
+        # -> _add_chunks_to_queue
+        # -> close
+        wait_for_write_lock()
 
     def close(self):
         if self._closed:
