@@ -11,7 +11,22 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
-#include <math.h>
+#include <stdint.h>
+#include <inttypes.h>
+
+#ifdef _WIN32
+#define _STDINT_H
+#endif
+#if !defined(__APPLE__)
+#include <malloc.h>
+#endif
+
+#include "x264lib.h"
+#include <x264.h>
+#include <libswscale/swscale.h>
+#include <libavcodec/avcodec.h>
+#include <libavutil/mem.h>
+
 
 //not honoured on MS Windows:
 #define MEMALIGN 1
@@ -21,26 +36,6 @@
 //(ie: when not supported by the library we build against)
 #define SUPPORT_CSC_MODES 1
 
-#ifdef _WIN32
-#include <malloc.h>
-#include "stdint.h"
-#include "inttypes.h"
-#else
-#include <stdint.h>
-#include <unistd.h>
-#endif
-
-#ifdef _WIN32
-typedef void x264_t;
-#define inline __inline
-#else
-#include <x264.h>
-#endif
-
-#include <libswscale/swscale.h>
-#include <libavcodec/avcodec.h>
-#include <libavutil/mem.h>
-#include "x264lib.h"
 
 //beware that these macros may evaluate a or b twice!
 //ie: do not use them for something like: MAX(i++, N)
@@ -62,6 +57,7 @@ struct x264lib_ctx {
 	// Encoding
 	x264_t *encoder;
 	struct SwsContext *rgb2yuv;
+	int use_swscale;
 
 	int speed;					//percentage 0-100
 	int quality;				//percentage 0-100
@@ -108,11 +104,10 @@ int get_encoder_speed(struct x264lib_ctx *ctx) {
 	return ctx->speed;
 }
 
-#ifndef _WIN32
 //Given a quality percentage (0 to 100),
 //return the x264 quality constant to use
 float get_x264_quality(int pct) {
-	return	roundf(50.0 - (MIN(100, MAX(0, pct)) * 49.0 / 100.0));
+	return	50.0f - (MIN(100, MAX(0, pct)) * 49.0f / 100.0f);
 }
 
 //Given a quality percentage (0 to 100),
@@ -169,7 +164,6 @@ int get_csc_format_for_x264_format(int i_csp)
 		return -1;
 	}
 }
-#endif
 
 //Given a csc colour sampling constant,
 //return our own generic csc constant (see codec_constants.py)
@@ -211,7 +205,6 @@ const char *DEFAULT_I444_PROFILE = PROFILE_HIGH444_PREDICTIVE;
 const int DEFAULT_I422_MIN_QUALITY = 80;
 const int DEFAULT_I444_MIN_QUALITY = 90;
 
-#ifndef _WIN32
 //Given a quality percentage (0 to 100)
 //return the profile to use
 //IMPORTANT: changes here must be reflected in get_x264_colour_sampling
@@ -232,9 +225,9 @@ const char *get_profile_for_quality(struct x264lib_ctx *ctx, int pct) {
 const char *get_valid_profile(const char* csc_mode, const char *profile, const char *profiles[], const char *default_profile)
 {
 	//printf("get_valid_profile(%s, %s, %p, %s)\n", csc_mode, profile, profiles, default_profile);
+	int i = 0;
 	if (profile==NULL)
 		return	default_profile;
-	int i = 0;
 	while (profiles[i]!=NULL)
 	{
 		if (strcmp(profiles[i], profile)==0) {
@@ -267,6 +260,7 @@ void configure_encoder(struct x264lib_ctx *ctx, int width, int height,
 		char *i420_profile, char *i422_profile, char *i444_profile)
 {
 	//printf("configure_encoder(%p, %i, %i, %i, %i, %i, %i, %s, %s, %s)\n", ctx, width, height, initial_quality, supports_csc_option, I422_quality, I444_quality, i420_profile, i422_profile, i444_profile);
+	ctx->use_swscale = 1;
 	ctx->width = width;
 	ctx->height = height;
 	if (initial_speed >= 0)
@@ -312,6 +306,7 @@ void configure_encoder(struct x264lib_ctx *ctx, int width, int height,
  */
 void do_init_encoder(struct x264lib_ctx *ctx)
 {
+	x264_param_t param;
 	ctx->colour_sampling = get_x264_colour_sampling(ctx, ctx->quality);
 	ctx->x264_quality = get_x264_quality(ctx->quality);
 	ctx->csc_format = get_csc_format_for_x264_format(ctx->colour_sampling);
@@ -321,7 +316,6 @@ void do_init_encoder(struct x264lib_ctx *ctx)
 	ctx->csc_algo = get_csc_algo_for_quality(ctx->quality);
 	//printf("do_init_encoder(%p, %i, %i, %i, %i) colour_sampling=%i, initial x264_quality=%f, initial profile=%s\n", ctx, ctx->width, ctx->height, ctx->quality, ctx->supports_csc_option, ctx->colour_sampling, ctx->x264_quality, ctx->profile);
 
-	x264_param_t param;
 	x264_param_default_preset(&param, ctx->preset, "zerolatency");
 	param.i_threads = 1;
 	param.i_width = ctx->width;
@@ -335,7 +329,8 @@ void do_init_encoder(struct x264lib_ctx *ctx)
 	param.b_open_gop = 1;			//allow open gop
 	x264_param_apply_profile(&param, ctx->profile);
 	ctx->encoder = x264_encoder_open(&param);
-	ctx->rgb2yuv = init_encoder_csc(ctx);
+	if (ctx->use_swscale)
+		ctx->rgb2yuv = init_encoder_csc(ctx);
 }
 
 struct x264lib_ctx *init_encoder(int width, int height,
@@ -346,6 +341,8 @@ struct x264lib_ctx *init_encoder(int width, int height,
         char *i420_profile, char *i422_profile, char *i444_profile)
 {
 	struct x264lib_ctx *ctx = malloc(sizeof(struct x264lib_ctx));
+	if (ctx==NULL)
+		return NULL;
 	memset(ctx, 0, sizeof(struct x264lib_ctx));
 	configure_encoder(ctx, width, height, \
 					initial_quality, initial_speed, \
@@ -375,32 +372,18 @@ void do_clean_encoder(struct x264lib_ctx *ctx)
 	}
 }
 
-#else
-struct x264lib_ctx *init_encoder(int width, int height,
-		int initial_quality, int initial_speed,
-		int supports_csc_option,
-		int I422_quality, int I444_quality,
-		int I422_min, int I444_min,
-        char *i420_profile, char *i422_profile, char *i444_profile)
-{
-	return NULL;
-}
 
-void clean_encoder(struct x264lib_ctx *ctx)
-{
-	return;
-}
-#endif
-
-int init_decoder_context(struct x264lib_ctx *ctx, int width, int height, int csc_fmt)
+int init_decoder_context(struct x264lib_ctx *ctx, int width, int height, int use_swscale, int csc_fmt)
 {
 	if (csc_fmt<0)
 		csc_fmt = PIX_FMT_YUV420P;
+	ctx->use_swscale = use_swscale;
 	ctx->width = width;
 	ctx->height = height;
 	ctx->csc_format = csc_fmt;
 	ctx->csc_algo = get_csc_algo_for_quality(100);
-	ctx->yuv2rgb = sws_getContext(ctx->width, ctx->height, ctx->csc_format, ctx->width, ctx->height, PIX_FMT_RGB24, ctx->csc_algo, NULL, NULL, NULL);
+	if (use_swscale)
+		ctx->yuv2rgb = sws_getContext(ctx->width, ctx->height, ctx->csc_format, ctx->width, ctx->height, PIX_FMT_RGB24, ctx->csc_algo, NULL, NULL, NULL);
 
 	avcodec_register_all();
 
@@ -410,6 +393,10 @@ int init_decoder_context(struct x264lib_ctx *ctx, int width, int height, int csc
 		return 1;
 	}
 	ctx->codec_ctx = avcodec_alloc_context3(ctx->codec);
+	if (!ctx->codec_ctx) {
+		fprintf(stderr, "failed to allocate codec context!\n");
+		return 1;
+	}
 	ctx->codec_ctx->width = ctx->width;
 	ctx->codec_ctx->height = ctx->height;
 	ctx->codec_ctx->pix_fmt = csc_fmt;
@@ -424,12 +411,14 @@ int init_decoder_context(struct x264lib_ctx *ctx, int width, int height, int csc
 	}
 	return 0;
 }
-struct x264lib_ctx *init_decoder(int width, int height, int csc_fmt)
+struct x264lib_ctx *init_decoder(int width, int height, int use_swscale, int csc_fmt)
 {
 	struct x264lib_ctx *ctx = malloc(sizeof(struct x264lib_ctx));
+	if (ctx==NULL)
+		return NULL;
 	memset(ctx, 0, sizeof(struct x264lib_ctx));
-	if (init_decoder_context(ctx, width, height, csc_fmt)) {
-		free(ctx);
+	if (init_decoder_context(ctx, width, height, use_swscale, csc_fmt)) {
+		clean_decoder(ctx);
 		return NULL;
 	}
 	return ctx;
@@ -437,6 +426,10 @@ struct x264lib_ctx *init_decoder(int width, int height, int csc_fmt)
 
 void do_clean_decoder(struct x264lib_ctx *ctx)
 {
+	if (ctx->frame) {
+		avcodec_free_frame(&ctx->frame);
+		ctx->frame = NULL;
+	}
 	if (ctx->codec_ctx) {
 		avcodec_close(ctx->codec_ctx);
 		av_free(ctx->codec_ctx);
@@ -446,7 +439,6 @@ void do_clean_decoder(struct x264lib_ctx *ctx)
 		sws_freeContext(ctx->yuv2rgb);
 		ctx->yuv2rgb = NULL;
 	}
-	avcodec_free_frame(&ctx->frame);
 }
 void clean_decoder(struct x264lib_ctx *ctx)
 {
@@ -454,13 +446,16 @@ void clean_decoder(struct x264lib_ctx *ctx)
 	free(ctx);
 }
 
-#ifndef _WIN32
+
 x264_picture_t *csc_image_rgb2yuv(struct x264lib_ctx *ctx, const uint8_t *in, int stride)
 {
+	x264_picture_t *pic_in = NULL;
 	if (!ctx->encoder || !ctx->rgb2yuv)
 		return NULL;
 
-	x264_picture_t *pic_in = malloc(sizeof(x264_picture_t));
+	pic_in = malloc(sizeof(x264_picture_t));
+	if (ctx==NULL)
+		return NULL;
 	x264_picture_alloc(pic_in, ctx->colour_sampling, ctx->width, ctx->height);
 
 	/* Colorspace conversion (RGB -> I4??) */
@@ -476,19 +471,21 @@ static void free_csc_image(x264_picture_t *image)
 
 int compress_image(struct x264lib_ctx *ctx, x264_picture_t *pic_in, uint8_t **out, int *outsz)
 {
+	x264_nal_t* nals = NULL;
+	int i_nals = 0;
+	x264_picture_t pic_out;
+	int frame_size = 0;
+
 	if (!ctx->encoder || !ctx->rgb2yuv) {
 		free_csc_image(pic_in);
 		*out = NULL;
 		*outsz = 0;
 		return 1;
 	}
-	x264_picture_t pic_out;
 
 	/* Encoding */
 	pic_in->i_pts = 1;
-	x264_nal_t* nals;
-	int i_nals;
-	int frame_size = x264_encoder_encode(ctx->encoder, &nals, &i_nals, pic_in, &pic_out);
+	frame_size = x264_encoder_encode(ctx->encoder, &nals, &i_nals, pic_in, &pic_out);
 	// Unconditional cleanup:
 	free_csc_image(pic_in);
 	if (frame_size < 0) {
@@ -502,16 +499,7 @@ int compress_image(struct x264lib_ctx *ctx, x264_picture_t *pic_in, uint8_t **ou
 	*outsz = frame_size;
 	return 0;
 }
-#else
-x264_picture_t* csc_image_rgb2yuv(struct x264lib_ctx *ctx, const uint8_t *in, int stride)
-{
-	return	NULL;
-}
-int compress_image(struct x264lib_ctx *ctx, x264_picture_t *pic_in, uint8_t **out, int *outsz)
-{
-	return 1;
-}
-#endif
+
 
 int csc_image_yuv2rgb(struct x264lib_ctx *ctx, uint8_t *in[3], const int stride[3], uint8_t **out, int *outsz, int *outstride)
 {
@@ -539,7 +527,7 @@ void set_decoder_csc_format(struct x264lib_ctx *ctx, int csc_fmt)
 	if (ctx->csc_format!=csc_fmt) {
 		//we need to re-initialize with the new format:
 		do_clean_decoder(ctx);
-		if (init_decoder_context(ctx, ctx->width, ctx->height, csc_fmt)) {
+		if (init_decoder_context(ctx, ctx->width, ctx->height, ctx->use_swscale, csc_fmt)) {
 			fprintf(stderr, "Failed to reconfigure decoder\n");
 		}
 	}
@@ -589,13 +577,12 @@ int decompress_image(struct x264lib_ctx *ctx, uint8_t *in, int size, uint8_t *(*
  * Change the speed of encoding (x264 preset).
  * @param percent: 100 for maximum ("ultrafast") with lowest compression, 0 for highest compression (slower)
  */
-#ifndef _WIN32
 void set_encoding_speed(struct x264lib_ctx *ctx, int pct)
 {
 	x264_param_t param;
+	int new_preset = 7-MAX(0, MIN(6, pct/16));
 	x264_encoder_parameters(ctx->encoder, &param);
 	ctx->speed = pct;
-	int new_preset = 7-MAX(0, MIN(6, pct/16.7));
 	if (new_preset==ctx->encoding_preset)
 		return;
 	//printf("set_encoding_speed(%i) old preset: %i=%s, new preset: %i=%s\n", pct, ctx->encoding_preset, x264_preset_names[ctx->encoding_preset], new_preset, x264_preset_names[new_preset]);
@@ -609,20 +596,14 @@ void set_encoding_speed(struct x264lib_ctx *ctx, int pct)
 	x264_param_apply_profile(&param, ctx->profile);
 	x264_encoder_reconfig(ctx->encoder, &param);
 }
-#else
-void set_encoding_speed(struct x264lib_ctx *ctx, int pct)
-{
-	;
-}
-#endif
 
 /**
  * Change the quality of encoding (x264 f_rf_constant).
  * @param percent: 100 for best quality, 0 for lowest quality.
  */
-#ifndef _WIN32
 void set_encoding_quality(struct x264lib_ctx *ctx, int pct)
 {
+	int old_csc_algo = ctx->csc_algo;
 	float new_quality = get_x264_quality(pct);
 	//printf("set_encoding_quality(%i) new_quality=%f, can csc=%i\n", pct, new_quality, ctx->supports_csc_option);
 	if (ctx->supports_csc_option) {
@@ -653,18 +634,11 @@ void set_encoding_quality(struct x264lib_ctx *ctx, int pct)
 		param.rc.f_rf_constant = new_quality;
 		x264_encoder_reconfig(ctx->encoder, &param);
 	}
-	int old_csc_algo = ctx->csc_algo;
 	ctx->csc_algo = get_csc_algo_for_quality(pct);
 	if (old_csc_algo!=ctx->csc_algo) {
 		ctx->rgb2yuv = init_encoder_csc(ctx);
 	}
 }
-#else
-void set_encoding_quality(struct x264lib_ctx *ctx, int pct)
-{
-	;
-}
-#endif
 
 
 void* xmemalign(size_t size)
